import os.path

import pytest

from uberspace_takeout.storage import LocalMoveStorage, Storage, TarStorage
from uberspace_takeout.compat import FileExistsError
from uberspace_takeout.compat import FileNotFoundError


@pytest.mark.parametrize('mode', [
    'takeout',
    'takein',
])
def test_storage_ctor(mode):
    s = Storage('dest.tar.gz', mode)
    assert s.destination == 'dest.tar.gz'
    assert s.mode == mode


def test_storage_ctor_mode():
    with pytest.raises(Exception) as ex:
        Storage('dest.tar.gz', 'foo')

    assert 'Invalid mode foo' in str(ex)


@pytest.fixture
def test_file(tmp_path):
    path = tmp_path / 'some_file.txt'

    with path.open('w') as f:
        f.write(u'some file text')

    return path


@pytest.fixture
def test_file2(tmp_path):
    path = tmp_path / 'some_file2.txt'

    with path.open('w') as f:
        f.write(u'some file text2')

    return path


@pytest.fixture
def test_dir(tmp_path):
    (tmp_path / 'some_dir/some_subdir').mkdir(parents=True)

    with (tmp_path / 'some_dir/file.txt').open('w') as f:
        f.write(u'some file text')

    with (tmp_path / 'some_dir/some_subdir/file2.txt').open('w') as f:
        f.write(u'some file text 2')

    return tmp_path / 'some_dir'


storages = [
    TarStorage,
    LocalMoveStorage,
]


@pytest.mark.parametrize('storage', storages)
def test_storage_text(storage, tmp_path):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_text('some text 1', 'simple_text.txt')
        s.store_text('some text 2', 'subdir/bla/simple_text.txt')

    with storage(tmp_path / 'test.tar.gz', 'takein') as s:
        assert s.unstore_text('simple_text.txt') == 'some text 1'
        assert s.unstore_text('subdir/bla/simple_text.txt') == 'some text 2'


@pytest.mark.parametrize('storage', storages)
def test_storage_file(storage, tmp_path, test_file, test_file2):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_file(test_file, 'simple_file.txt')
        s.store_file(test_file2, 'subdir/bla/simple_file.txt')

    with storage(tmp_path / 'test.tar.gz', 'takein') as s:
        assert s.unstore_text('simple_file.txt') == 'some file text'
        # read twice to check hat we don't break the tarinfo objects
        assert s.unstore_text('subdir/bla/simple_file.txt') == 'some file text2'
        assert s.unstore_text('subdir/bla/simple_file.txt') == 'some file text2'

        s.unstore_file('simple_file.txt', tmp_path / 'extracted.txt')

        with (tmp_path / 'extracted.txt').open() as f:
            assert f.read() == 'some file text'


@pytest.mark.parametrize('storage', storages)
def test_storage_file_eexists(storage, tmp_path, test_file):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_file(test_file, 'simple_file.txt')

        test_file.touch()

        with pytest.raises(FileExistsError):
            s.store_file(test_file, 'simple_file.txt')


@pytest.mark.parametrize('storage', storages)
def test_storage_file_noent(storage, tmp_path, test_file):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        pass

    with storage(tmp_path / 'test.tar.gz', 'takein') as s:
        with pytest.raises(FileNotFoundError):
            s.unstore_file('simple_file.txt', tmp_path / 'somefile')


@pytest.mark.parametrize('storage', storages)
def test_storage_directory(storage, tmp_path, test_dir):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_directory(test_dir, 'dir')

    with storage(tmp_path / 'test.tar.gz', 'takein') as s:
        assert s.unstore_text('dir/some_subdir/file2.txt') == 'some file text 2'

        s.unstore_directory('dir', tmp_path / 'new_dir')

        with (tmp_path / 'new_dir' / 'file.txt').open() as f:
            assert f.read() == 'some file text'

        with (tmp_path / 'new_dir' / 'some_subdir/file2.txt').open() as f:
            assert f.read() == 'some file text 2'


@pytest.mark.parametrize('storage', storages)
def test_storage_directory_eexists(storage, tmp_path, test_dir):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_directory(test_dir, 'dir')

        test_dir.mkdir(exist_ok=True)

        with pytest.raises(FileExistsError):
            s.store_directory(test_dir, 'dir')


@pytest.mark.parametrize('storage', storages)
def test_storage_directory_noent(storage, tmp_path):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        pass

    with storage(tmp_path / 'test.tar.gz', 'takein') as s:
        with pytest.raises(FileNotFoundError):
            s.unstore_directory('dir', tmp_path / 'new_dir')


@pytest.mark.parametrize('storage', storages)
def test_storage_list_files(storage, tmp_path, test_dir):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_directory(test_dir, 'dir')
        assert sorted(s.list_files('dir')) == ['file.txt', 'some_subdir']
        assert sorted(s.list_files('dir/some_subdir')) == ['file2.txt']


@pytest.mark.parametrize('storage', storages)
def test_storage_list_files_noent(storage, tmp_path, test_dir):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        pass

    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        with pytest.raises(FileNotFoundError):
            s.list_files('dir')


@pytest.mark.parametrize('storage', storages)
def test_storage_slashes(storage, tmp_path, test_file):
    with storage(tmp_path / 'test.tar.gz', 'takeout') as s:
        s.store_text('some text 3', '/subdir/bla/simple_text2.txt')
        s.store_file(test_file, '/subdir/bla/simple_file2.txt')

    with storage(tmp_path / 'test.tar.gz', 'takein') as s:
        assert s.unstore_text('/subdir/bla/simple_text2.txt') == 'some text 3'
        assert s.unstore_text('/subdir/bla/simple_file2.txt') == 'some file text'
        assert s.unstore_text('subdir/bla/simple_text2.txt') == 'some text 3'
        assert s.unstore_text('subdir/bla/simple_file2.txt') == 'some file text'


def test_localmovestorage_file_removed(tmp_path, test_file):
    with LocalMoveStorage(tmp_path / 'test.tar.gz', 'takeout') as s:
        assert test_file.exists()
        s.store_file(test_file, 'simple_file.txt')
        assert not test_file.exists()


def test_localmovestorage_directory_removed(tmp_path, test_dir):
    with LocalMoveStorage(tmp_path / 'test.tar.gz', 'takeout') as s:
        assert test_dir.exists()
        s.store_directory(test_dir, 'dir')
        assert not test_dir.exists()



import datetime
import errno
import os
import tarfile

from .compat import BytesIO
from .compat import FileExistsError
from .compat import FileNotFoundError


class Storage():
    def __init__(self, destination, mode):
        if mode not in ('takein', 'takeout'):
            raise Exception('Invalid mode {}, expected "takein" or "takeout".'.format(mode))

        self.destination = str(destination)
        self.mode = mode

    def __enter__(self):
        raise NotImplementedError()

    def __exit__(self, exception_type, exception_value, traceback):
        raise NotImplementedError()

    def list_files(self, storage_path):
        raise NotImplementedError()

    def store_text(self, content, storage_path):
        raise NotImplementedError()

    def unstore_text(self, storage_path):
        raise NotImplementedError()

    def store_file(self, system_path, storage_path):
        raise NotImplementedError()

    def unstore_file(self, storage_path, system_path):
        raise NotImplementedError()

    def store_directory(self, system_path, storage_path):
        return self.store_file(system_path, storage_path)

    def unstore_directory(self, storage_path, system_path):
        return self.unstore_file(storage_path, system_path)


class TarStorage(Storage):

    def __enter__(self):
        mode = 'w:bz2' if self.mode == 'takeout' else 'r:bz2'
        self.tar = tarfile.open(self.destination, mode)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.tar.close()

    def _check_member_type(self, member):
        if member.type not in (tarfile.REGTYPE, tarfile.SYMTYPE, tarfile.DIRTYPE):
            raise Exception(
                'tar member has illegal type: {}. '
                'Must be tarfile.REGTYPE/file, SYMTYPE/symlink or DIRTYPE/directory, '
                'but is {}'.format(member.name, member.type)
            )

    def clone_tarinfo(self, tarinfo):
        # "clone" the object so we don't modify names inside the tar
        tarinfo2 = tarfile.TarInfo()
        try:
            attributes = tuple(tarinfo.get_info().keys())
        except TypeError:
            # Python 2.7 needs additional arguments for encoding
            attributes = tuple(tarinfo.get_info(encoding='utf-8', errors='strict').keys())
        for attr in attributes + ('offset', 'offset_data'):
            setattr(tarinfo2, attr, getattr(tarinfo, attr))
        return tarinfo2

    def get_members_in(self, directory):
        directory = directory.rstrip('/') + '/'

        for m in self.tar.getmembers():
            if '..' in m.name:
                raise Exception('tar member has illegal name (contains ".."): ' + m.name)
            if m.name.startswith('/'):
                raise Exception('tar member has illegal name (starts with "/"): ' + m.name)
            if m.name.startswith('./'):
                raise Exception('tar member has illegal name (starts with "./"): ' + m.name)

            self._check_member_type(m)

            if m.name.startswith(directory):
                m = self.clone_tarinfo(m)
                # files might be stored as /www/domain.com/something.html, but need to be extracted
                # as domain.com/something.html.
                m.name = m.name[len(directory):]
                yield m

    def has_member(self, path):
        for m in self.tar.getmembers():
            self._check_member_type(m)

            if m.name == path:
                return True

        return False

    def get_member(self, path):
        matching = []

        for m in self.tar.getmembers():
            self._check_member_type(m)

            if m.name == path:
                matching.append(m)

        if len(matching) == 0:
            raise FileNotFoundError()
        if len(matching) > 1:
            raise Exception('There are {} files matching the path {}. Expected only one.'.format(len(matching), path))

        return self.clone_tarinfo(matching[0])

    @classmethod
    def _len(cls, f):
        old_position = f.tell()
        f.seek(0, os.SEEK_END)
        length = f.tell()
        f.seek(old_position)
        return length

    def list_files(self, storage_path):
        members = list(self.get_members_in(storage_path))
        if not members:
            raise FileNotFoundError()
        return [m.name for m in members if '/' not in m.name]

    def store_text(self, content, storage_path):
        storage_path = str(storage_path).lstrip('/')
        content = BytesIO(content.encode('utf-8'))
        info = tarfile.TarInfo(storage_path)
        info.size = self._len(content)
        info.mtime = int(datetime.datetime.now().strftime("%s"))
        self.tar.addfile(info, content)

    def unstore_text(self, storage_path):
        storage_path = str(storage_path).lstrip('/')
        if not self.has_member(storage_path):
            raise FileNotFoundError()
        return self.tar.extractfile(storage_path).read().decode('utf-8')

    def store_file(self, system_path, storage_path):
        storage_path = str(storage_path).lstrip('/')
        if self.has_member(storage_path):
            raise FileExistsError()
        self.tar.add(str(system_path), storage_path)

    def unstore_directory(self, storage_path, system_path):
        system_path = str(system_path)  # for Python 2.7
        storage_path = str(storage_path).lstrip('/')
        members = list(self.get_members_in(storage_path))
        if not members:
            raise FileNotFoundError()
        self.tar.extractall(system_path, members)

    def unstore_file(self, storage_path, system_path):
        system_path = str(system_path)  # for Python 2.7
        storage_path = str(storage_path).lstrip('/')
        member = self.get_member(storage_path)
        member.name = os.path.basename(system_path)
        self.tar.extractall(os.path.dirname(system_path), [member])


class LocalMoveStorage(Storage):

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass

    def _storage_path(self, storage_path):
        storage_path = str(storage_path).lstrip('/')
        return self.destination + '/' + storage_path

    def _mkdir_p(self, path):
        if not path:
            return

        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    def list_files(self, storage_path):
        storage_path = self._storage_path(storage_path)
        if not os.path.exists(storage_path):
            raise FileNotFoundError()
        return os.listdir(storage_path)

    def store_text(self, content, storage_path):
        storage_path = self._storage_path(storage_path)
        if os.path.exists(storage_path):
            raise FileExistsError()
        self._mkdir_p(os.path.dirname(storage_path))
        with open(storage_path, 'w') as f:
            f.write(content)

    def unstore_text(self, storage_path):
        with open(self._storage_path(storage_path)) as f:
            return f.read()

    def store_file(self, system_path, storage_path):
        system_path = str(system_path)  # for Python 2.7
        storage_path = self._storage_path(storage_path)
        if os.path.exists(storage_path):
            raise FileExistsError()
        self._mkdir_p(os.path.dirname(storage_path))
        os.rename(system_path, storage_path)

    def unstore_file(self, storage_path, system_path):
        system_path = str(system_path)  # for Python 2.7
        storage_path = self._storage_path(storage_path)
        if not os.path.exists(storage_path):
            raise FileNotFoundError()
        self._mkdir_p(os.path.dirname(system_path))
        os.rename(storage_path, system_path)



try:
    from BytesIO import BytesIO
except ImportError:
    from io import BytesIO  # noqa

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # noqa

try:
    FileExistsError = FileExistsError
    FileNotFoundError = FileNotFoundError
except NameError:
    FileExistsError = IOError
    FileNotFoundError = IOError



import os
import shutil

import pytest
from pyfakefs.fake_filesystem_unittest import Pause

from uberspace_takeout import Takeout
from uberspace_takeout.compat import FileNotFoundError
from uberspace_takeout.compat import Path


# NOTE: there is a lot of `str(PosixPath)` casting going on below, this might
# be better solved with using `pathlib.Path` instead of `os` functions...


def prefix_root(prefix):
    return Path(__file__).parent / 'uberspaces' / prefix


def populate_root(fs, prefix):
    outside_root = prefix_root(prefix)

    with Pause(fs):
        for dir in os.listdir(str(outside_root)):
            if not os.path.isdir(str(outside_root / dir)):
                raise NotImplementedError("currently only directories are supported at root level")
            if dir == 'commands':
                continue

            try:
                fs.remove_object('/' + dir)
            except FileNotFoundError:
                pass

            fs.add_real_directory(str(outside_root / dir), lazy_read=False, read_only=False, target_path='/' + dir)


def clean_root(skip_dirs=['tmp', 'etc']):
    for path in os.listdir('/'):
        if path not in skip_dirs:
            shutil.rmtree('/' + path)

    Path('/home').mkdir()
    Path('/var/www/virtual').mkdir(parents=True)

    assert not os.listdir('/home')
    assert not os.listdir('/var/www/virtual')


@pytest.fixture
def mock_run_command(fs, mocker):
    class Commands():
        def __init__(self, *args, **kwargs):
            self.called = {}
            self.commands = {}

            self._mock()

        def clear(self):
            self.called.clear()
            self.commands.clear()

        def add_prefix_commands(self, prefix):
            commands = prefix_root(prefix) / 'commands'

            with Pause(fs):
                for cmd in os.listdir(str(commands)):
                    with open(str(commands / cmd)) as f:
                        self.add_command(cmd, f.read())

        def add_command(self, command, output=""):
            if command in self.commands:
                raise Exception("Command '{}' is already defined.".format(command))
            if isinstance(command, list):
                command = " ".join(command)

            self.commands[command] = output

        def assert_called(self, command, input_text=None):
            if isinstance(command, list):
                command = " ".join(command)

            assert command in self.called

            if input_text:
                assert self.called[command] == input_text

            del self.called[command]

        def assert_no_unexpected(self):
            assert not self.called, "unexpected commands executed"

        def _mock(self):
            def _run_command(_, cmd, input_text=None, *args, **kwargs):
                cmd_str = " ".join(cmd)
                self.called[cmd_str] = input_text

                if cmd_str in self.commands:
                    output = self.commands[cmd_str]
                    return [l.rstrip() for l in output.split('\n') if l.rstrip()]
                else:
                    return []

            mocker.patch('uberspace_takeout.items.base.TakeoutItem.run_command', _run_command)

    return Commands()


def content(path):
    with open(str(path)) as f:
        return f.read()


def assert_in_file(path, text):
    assert text in content(path)


def assert_files_equal(path1, path2):
    assert content(path1) == content(path2)


def assert_file_unchanged(path, fs, prefix):
    with Pause(fs):
        original = content(prefix_root(prefix) / path.lstrip('/'))

    assert original == content(path)


def test_takeout_u6_to_u6(fs, mock_run_command):
    populate_root(fs, 'u6/isabell')
    mock_run_command.add_prefix_commands('u6/isabell')

    takeout = Takeout(hostname='andromeda.uberspace.de')

    takeout.takeout('/tmp/test.tar.gz', 'isabell')

    clean_root()

    mock_run_command.clear()

    takeout.takein('/tmp/test.tar.gz', 'isabell')

    mock_run_command.assert_called("mysql --defaults-group-suffix= -e SET PASSWORD = PASSWORD('Lei4eengekae3iet4Ies')")
    assert_in_file('/home/isabell/.my.cnf', 'Lei4eengekae3iet4Ies')
    mock_run_command.assert_called("mysql --defaults-group-suffix=readonly -e SET PASSWORD = PASSWORD('eeruaSooch6iereequoo')")
    assert_in_file('/home/isabell/.my.cnf', 'eeruaSooch6iereequoo')

    mock_run_command.assert_called("uberspace-add-domain -w -d *.example.com")
    mock_run_command.assert_called("uberspace-add-domain -w -d example.com")
    mock_run_command.assert_called("uberspace-add-domain -w -d foo.example.com")
    mock_run_command.assert_called("uberspace-add-domain -m -d mail.example.com")

    mock_run_command.assert_called("crontab -", "@daily echo good morning\n")

    mock_run_command.assert_no_unexpected()

    assert_file_unchanged('/var/www/virtual/isabell/html/index.html', fs, 'u6/isabell')
    assert_file_unchanged('/var/www/virtual/isabell/html/blog/index.html', fs, 'u6/isabell')
    assert os.path.islink('/home/isabell/html')
    assert_file_unchanged('/home/isabell/html/index.html', fs, 'u6/isabell')
    assert_file_unchanged('/home/isabell/Maildir/cur/mail-888', fs, 'u6/isabell')


def test_takeout_u6_to_u7(fs, mock_run_command):
    populate_root(fs, 'u6/isabell')
    mock_run_command.add_prefix_commands('u6/isabell')

    takeout = Takeout(hostname='andromeda.uberspace.de')

    takeout.takeout('/tmp/test.tar.gz', 'isabell')

    clean_root()
    mock_run_command.clear()
    populate_root(fs, 'u7/empty')

    takeout.takein('/tmp/test.tar.gz', 'isabell')

    mock_run_command.assert_called("mysql --defaults-group-suffix= -e SET PASSWORD = PASSWORD('Lei4eengekae3iet4Ies')")
    assert_in_file('/home/isabell/.my.cnf', 'Lei4eengekae3iet4Ies')
    mock_run_command.assert_called("mysql --defaults-group-suffix=readonly -e SET PASSWORD = PASSWORD('eeruaSooch6iereequoo')")
    assert_in_file('/home/isabell/.my.cnf', 'eeruaSooch6iereequoo')

    mock_run_command.assert_called("uberspace web domain add example.com")
    mock_run_command.assert_called("uberspace web domain add foo.example.com")
    mock_run_command.assert_called("uberspace mail domain add mail.example.com")

    mock_run_command.assert_called("crontab -", "@daily echo good morning\n")

    mock_run_command.assert_no_unexpected()

    assert_file_unchanged('/var/www/virtual/isabell/html/index.html', fs, 'u6/isabell')
    assert_file_unchanged('/var/www/virtual/isabell/html/blog/index.html', fs, 'u6/isabell')
    assert os.path.islink('/home/isabell/html')
    assert_file_unchanged('/home/isabell/html/index.html', fs, 'u6/isabell')
    assert_file_unchanged('/home/isabell/Maildir/cur/mail-888', fs, 'u6/isabell')



import re

from .base import TakeoutItem, UberspaceVersionMixin
from ..compat import FileNotFoundError


class U7Mixin(UberspaceVersionMixin):
    uberspace_version = 7


class DomainItem(U7Mixin, TakeoutItem):
    area = None

    def takeout(self):
        domains = self.run_uberspace(self.area, 'domain', 'list')
        domains = set(domains) - {
            self.username + '.uber.space',
            self.username + '.' + self.hostname
        }
        self.storage.store_text('\n'.join(domains), self.storage_path)

    def takein(self):
        text = self.storage.unstore_text(self.storage_path)
        for domain in (d for d in text.split('\n') if d):
            if domain.startswith('*.'):
                print("cannot add wildcard domain on: " + domain)
                continue
            self.run_uberspace(self.area, 'domain', 'add', domain)


class WebDomains(DomainItem):
    description = 'Web Domains'
    area = 'web'
    storage_path = 'conf/domains-web'


class MailDomains(DomainItem):
    description = 'Mail Domains'
    area = 'mail'
    storage_path = 'conf/domains-mail'


class FlagItem(U7Mixin, TakeoutItem):
    """
    A flag like "is the spamfilter enabled?". It provide a status/enable/disable
    interface via a uberspace sub-command. Provide the uberspace command without
    leading "uberspace" in `cmd` (e.g. ['web', 'log', 'access']).
    """

    def takeout(self):
        out = self.run_uberspace(*(self.cmd + ['status']))
        if 'enabled' in out:
            status = 'enable'
        else:
            status = 'disable'

        self.storage.store_text(status, self.storage_path)

    def takein(self):
        try:
            data = self.storage.unstore_text(self.storage_path)
        except FileNotFoundError:
            return

        if data not in ('enable', 'disable'):
            raise Exception(
                'invalid "uberspace {}" value: {}, expected "enabled" or "disabled".'.format(
                    ' '.join(self.cmd),
                    data,
                )
            )

        self.run_uberspace(*(self.cmd + [data]))


class AccessLogItem(FlagItem):
    description = 'Setting: Access-Log'
    cmd = ['web', 'log', 'access']
    storage_path = 'conf/log-access'


class ApacheErrorLogItem(FlagItem):
    description = 'Setting: Apache-Error-Log'
    cmd = ['web', 'log', 'apache_error']
    storage_path = 'conf/log-apache_error'


class PhpErrorLogItem(FlagItem):
    description = 'Setting: PHP-Error-Log'
    cmd = ['web', 'log', 'php_error']
    storage_path = 'conf/log-php_error'


class SpamfilterLogItem(FlagItem):
    description = 'Setting: Spamfilter'
    cmd = ['mail', 'spamfilter']
    storage_path = 'conf/spamfilter-enabled'


class ToolVersions(U7Mixin, TakeoutItem):
    description = 'Setting: Tool Versions'

    def takeout(self):
        tools = self.run_uberspace('tools', 'version', 'list')

        for tool in tools:
            tool = tool.lstrip('- ')
            out = self.run_uberspace('tools', 'version', 'show', tool)
            version = re.search(r"'([0-9\.]+)'", out[0]).groups()[0]
            self.storage.store_text(version, 'conf/tool-version/' + tool)

    def takein(self):
        try:
            tools = self.storage.list_files('conf/tool-versions')
        except FileNotFoundError:
            return

        for tool in tools:
            version = self.storage.unstore_text('conf/tool-versions/' + tool)
            self.run_uberspace('tools', 'version', 'use', tool, version)



import re

from .base import TakeoutItem, UberspaceVersionMixin


class U6Mixin(UberspaceVersionMixin):
    uberspace_version = 6


class DomainItem(U6Mixin, TakeoutItem):
    flag = None
    storage_path = None

    def takeout(self):
        domains = self.run_command(['uberspace-list-domains', self.flag])
        domains = set(domains) - {
            self.username + '.' + self.hostname,
            '*.' + self.username + '.' + self.hostname,
        }
        text = '\n'.join(domains)
        self.storage.store_text(text, self.storage_path)

    def takein(self):
        text = self.storage.unstore_text(self.storage_path)

        for domain in (d for d in text.split('\n') if d):
            self.run_command(['uberspace-add-domain', self.flag, '-d', domain])


class WebDomains(DomainItem):
    description = 'Web Domains'
    flag = '-w'
    storage_path = 'conf/domains-web'


class MailDomains(DomainItem):
    description = 'Mail Domains'
    flag = '-m'
    storage_path = 'conf/domains-mail'


class ToolVersion(TakeoutItem):
    storage_path = None
    version_regex = None
    software_command = None

    def _clean_version(self, version):
        return version

    def takeout(self):
        output = self.run_command(self.software_command)
        version = re.search(self.version_regex, output).groups()[0]
        version = self._clean_version(version)
        self.storage.store_text(version, self.storage_path)

    def takein(self):
        pass
        # TODO: print warning


class NodeVersion(ToolVersion):
    storage_path = 'conf/tool-version/node'
    version_regex = r'^v([0-9]+)'
    software_command = 'node --version'


class RubyVersion(ToolVersion):
    storage_path = 'conf/tool-version/ruby'
    version_regex = r'^ruby ([0-9]+\.[0-9]+)'
    software_command = 'ruby --version'


class PhpVersion(ToolVersion):
    storage_path = 'conf/tool-version/php'
    version_regex = r'^PHP ([0-9]+\.[0-9]+)'
    software_command = 'php --version'



import configparser

from .base import PathItem, TakeoutItem


class TakeoutMarker(TakeoutItem):
    description = 'Takeout Marker (internal)'

    def takeout(self):
        self.storage.store_text('uberspace_takeout', '.uberspace_takeout')

    def takein(self):
        content = self.storage.unstore_text('.uberspace_takeout')

        if content != 'uberspace_takeout':
            raise Exception('input is not a takeout.')


class Homedir(PathItem):
    description = 'Homedirectory'
    storage_path = 'home/'

    def path(self):
        return '/home/' + self.username


class Www(PathItem):
    description = 'Documentroot'
    storage_path = 'www/'

    def path(self):
        return '/var/www/virtual/' + self.username


class Cronjobs(TakeoutItem):
    description = 'Cronjobs'

    def takeout(self):
        cronjobs = self.run_command(['crontab', '-l'])

        if len(cronjobs) == 1 and cronjobs[0].startswith('no crontab for'):
            text = ''
        else:
            text = '\n'.join(cronjobs) + '\n'

        self.storage.store_text(text, 'conf/cronjobs')

    def takein(self):
        text = self.storage.unstore_text('conf/cronjobs')
        self.run_command(['crontab', '-'], input_text=text)


class MySQLPassword(TakeoutItem):
    description = 'MySQL password'

    @property
    def _my_cnf_path(self):
        return '/home/' + self.username + '/.my.cnf'

    def _open_my_cnf(self, section):
        config = configparser.ConfigParser()
        config.read(self._my_cnf_path)
        return config

    def _read_my_cnf_password(self, section):
        raw_pw = self._open_my_cnf(section)[section]['password']
        return raw_pw.partition('#')[0].strip()

    def _write_my_cnf_password(self, section, password):
        config = self._open_my_cnf(section)
        config[section]['password'] = password

        with open(self._my_cnf_path, 'w') as f:
            config.write(f)

    def _set_password(self, suffix):
        password = self.storage.unstore_text('conf/mysql-password-client' + suffix)
        self.run_command(['mysql', '--defaults-group-suffix=' + suffix, '-e', "SET PASSWORD = PASSWORD('" + password + "')"])
        self._write_my_cnf_password('client' + suffix, password)

    def takeout(self):
        self.storage.store_text(self._read_my_cnf_password('client'), 'conf/mysql-password-client')
        self.storage.store_text(self._read_my_cnf_password('clientreadonly'), 'conf/mysql-password-clientreadonly')

    def takein(self):
        self._set_password('')
        self._set_password('readonly')



from __future__ import print_function
import sys

try:
    from setuptools import setup
except ImportError:
    print("This project needs setuptools.", file=sys.stderr)
    print("Please install it using your package-manager or pip.", file=sys.stderr)
    sys.exit(1)

setup(
    name='uberspace_takeout',
    version='0.0.2',
    description='',
    author='uberspace.de',
    author_email='hallo@uberspace.de',
    url='https://github.com/uberspace/takeout',
    packages=[
      'uberspace_takeout',
      'uberspace_takeout.items',
    ],
    extras_require={
      'test': [
        'pyfakefs>=3.6',
        'pytest-mock',
      ],
      'py27': [
          'configparser',
          'pathlib2',
      ]
    },
    entry_points={
      'console_scripts': ['uberspace-takeout=uberspace_takeout.__main__:main'],
    },
    classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Developers',
      'Intended Audience :: Information Technology',
      'Intended Audience :: System Administrators',
      'Topic :: System :: Systems Administration',
      'Topic :: Security',
      'Topic :: Utilities',
      'Natural Language :: English',
      'Operating System :: POSIX :: Linux',
      'Programming Language :: Python :: 2.7',
    ],
    zip_safe=True,
)



import argparse
import datetime
import getpass
import sys

from . import Takeout


def main():
    username = getpass.getuser()
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
    default_tar_path = 'takeout_{}_{}.tar.bz2'.format(username, timestamp)

    p = argparse.ArgumentParser()
    p.add_argument('action', choices=['takeout', 'takein', 'items'])
    p.add_argument('--username', default=username)
    p.add_argument('--skip-item', action='append', default=[])
    p.add_argument('--tar-file', default=default_tar_path)
    args = p.parse_args()

    tar_path = args.tar_file

    if args.action == 'takeout':
        if tar_path == '-':
            tar_path = '/dev/stdout'
            sys.stdout = sys.stderr

        print('writing ' + tar_path)
        Takeout().takeout(tar_path, args.username, args.skip_item)
    elif args.action == 'takein':
        if tar_path == '-':
            tar_path = '/dev/stdin'

        print('reading ' + tar_path)
        Takeout().takein(tar_path, args.username, args.skip_item)
    elif args.action == 'items':
        print('\n'.join(i.__name__.ljust(25, ' ') + i.description for i in Takeout.takeout_menu))
    else:
        raise NotImplementedError()


if __name__ == '__main__':
    main()
